//--------------------------------- memalloc.h --------------------------------
// Functions for dynamic memory allocation of vectors and matrices.
// Part of the numxlib numerics library. Author: Titus Beu, 2019
//------------------------------------------------------------------------------
#pragma once
#include <stdlib.h>
#include <stdio.h>
#include <complex>

using namespace std;
typedef complex<double> dcmplx;

//==============================================================================
double *Vector(int imin, int imax)
//------------------------------------------------------------------------------
// Allocates a double vector with indices in the range [imin,imax]
//------------------------------------------------------------------------------
{
   double *p;
                                          // assign block start to array pointer
   p = (double*)malloc((size_t)((imax - imin + 1) * sizeof(double)));
   if (!p) { printf("Vector: allocation error !\n"); exit(1);
   }
   return p - imin;                                         // adjust for offset
}

//==============================================================================
void FreeVector(double *p, int imin)
//------------------------------------------------------------------------------
// Deallocates a double vector allocated with Vector, with offset imin
//------------------------------------------------------------------------------
{
   free((void*)(p + imin));                              // compensate for offset
}

//==============================================================================
double **Matrix(int imin, int imax, int jmin, int jmax)
//------------------------------------------------------------------------------
// Allocates a double matrix, with row and column indices in the range
// [imin,imax] x [jmin,jmax]
//------------------------------------------------------------------------------
{                                                 // numbers of rows and columns
   int i, ni = imax - imin + 1, nj = jmax - jmin + 1;
   double **p;
                                               // allocate array of row pointers
   p = (double**)malloc((size_t)(ni * sizeof(double*)));
   if (!p) {
      printf("Matrix: level 1 allocation error !\n");
      exit(1);
   }
   p -= imin;                                           // adjust for row offset
                                        // assign block start to 1st row pointer
   p[imin] = (double*)malloc((size_t)(ni*nj * sizeof(double)));
   if (!p[imin]) {
      printf("Matrix: level 2 allocation error !\n");
      exit(2);
   }
   p[imin] -= jmin;                  // adjust 1st row pointer for column offset
                                     // define row pointers spaced by row length
   for (i = imin + 1; i <= imax; i++) p[i] = p[i - 1] + nj;

   return p;
}

//==============================================================================
void FreeMatrix(double **p, int imin, int jmin)
//------------------------------------------------------------------------------
// Deallocates a double matrix allocated with Matrix, with row and column
// offsets imin and jmin
//------------------------------------------------------------------------------
{
   free((void*)(p[imin] + jmin));                            // deallocate block
   free((void*)(p + imin));                  // deallocate array of row pointers
}

//==============================================================================
int *IVector(int imin, int imax)
//------------------------------------------------------------------------------
// Allocates an int vector with indices in the range [imin,imax]
//------------------------------------------------------------------------------
{
   int *p;
                                          // assign block start to array pointer
   p = (int*)malloc((size_t)((imax - imin + 1) * sizeof(int)));
   if (!p) {
      printf("IVector: allocation error !\n");
      exit(1);
   }
   return p - imin;                                         // adjust for offset
}

//==============================================================================
void FreeIVector(int *p, int imin)
//------------------------------------------------------------------------------
// Deallocates an int vector allocated with IVector, with offset imin
//------------------------------------------------------------------------------
{
   free((void*)(p + imin));                             // compensate for offset
}

//==============================================================================
int **IMatrix(int imin, int imax, int jmin, int jmax)
//------------------------------------------------------------------------------
// Allocates an int matrix, with row and column indices in the range
// [imin,imax] x [jmin,jmax]
//------------------------------------------------------------------------------
{                                                 // numbers of rows and columns
   int i, ni = imax - imin + 1, nj = jmax - jmin + 1;
   int **p;
                                               // allocate array of row pointers
   p = (int**)malloc((size_t)(ni * sizeof(int*)));
   if (!p) {
      printf("IMatrix: level 1 allocation error !\n");
      exit(1);
   }
   p -= imin;                                           // adjust for row offset
                                        // assign block start to 1st row pointer
   p[imin] = (int*)malloc((size_t)(ni*nj * sizeof(int)));
   if (!p[imin]) {
      printf("IMatrix: level 2 allocation error !\n");
      exit(2);
   }
   p[imin] -= jmin;                  // adjust 1st row pointer for column offset
                                     // define row pointers spaced by row length
   for (i = imin + 1; i <= imax; i++) p[i] = p[i - 1] + nj;

   return p;
}

//==============================================================================
void FreeIMatrix(int **p, int imin, int jmin)
//------------------------------------------------------------------------------
// Deallocates an int matrix allocated with IMatrix, with row and column
// offsets imin and jmin
//------------------------------------------------------------------------------
{
   free((void*)(p[imin] + jmin));                            // deallocate block
   free((void*)(p + imin));                  // deallocate array of row pointers
}

//==============================================================================
long long *LVector(int imin, int imax)
//------------------------------------------------------------------------------
// Allocates a long long int vector with indices in the range [imin,imax]
//------------------------------------------------------------------------------
{
   long long *p;
                                          // assign block start to array pointer
   p = (long long*)malloc((size_t)((imax - imin + 1) * sizeof(long long)));
   if (!p) {
      printf("LVector: allocation error !\n");
      exit(1);
   }
   return p - imin;                                         // adjust for offset
}

//==============================================================================
void FreeLVector(long long *p, int imin)
//------------------------------------------------------------------------------
// Deallocates a long long int vector allocated with LVector, with offset imin
//------------------------------------------------------------------------------
{
   free((void*)(p + imin));                             // compensate for offset
}

//==============================================================================
long long **LMatrix(int imin, int imax, int jmin, int jmax)
//------------------------------------------------------------------------------
// Allocates a long long int matrix, with row and column indices in the range
// [imin,imax] x [jmin,jmax]
//------------------------------------------------------------------------------
{                                                 // numbers of rows and columns
   int i, ni = imax - imin + 1, nj = jmax - jmin + 1;
   long long **p;
                                               // allocate array of row pointers
   p = (long long**)malloc((size_t)(ni * sizeof(long long*)));
   if (!p) {
      printf("LMatrix: level 1 allocation error !\n");
      exit(1);
   }
   p -= imin;                                           // adjust for row offset
                                        // assign block start to 1st row pointer
   p[imin] = (long long*)malloc((size_t)(ni*nj * sizeof(long long)));
   if (!p[imin]) {
      printf("LMatrix: level 2 allocation error !\n");
      exit(2);
   }
   p[imin] -= jmin;                  // adjust 1st row pointer for column offset
                                     // define row pointers spaced by row length
   for (i = imin + 1; i <= imax; i++) p[i] = p[i - 1] + nj;

   return p;
}

//==============================================================================
void FreeLMatrix(long long **p, int imin, int jmin)
//------------------------------------------------------------------------------
// Deallocates a long long int matrix allocated with LMatrix, with row and
// column offsets imin and jmin
//------------------------------------------------------------------------------
{
   free((void*)(p[imin] + jmin));                            // deallocate block
   free((void*)(p + imin));                  // deallocate array of row pointers
}

//==============================================================================
dcmplx *CVector(int imin, int imax)
//------------------------------------------------------------------------------
// Allocates a complex vector with indices in the range [imin,imax]
//------------------------------------------------------------------------------
{
   dcmplx *p;
                                          // assign block start to array pointer
   p = (dcmplx*)malloc((size_t)((imax - imin + 1) * 2 * sizeof(double)));
   if (!p) {
      printf("CVector: allocation error !\n");
      exit(1);
   }
   return p - imin;                                         // adjust for offset
}

//==============================================================================
void FreeCVector(dcmplx *p, int imin)
//------------------------------------------------------------------------------
// Deallocates a complex vector allocated with CVector, with offset imin
//------------------------------------------------------------------------------
{
   free((void*)(p + imin));                              // compensate for offset
}

//==============================================================================
dcmplx **CMatrix(int imin, int imax, int jmin, int jmax)
//------------------------------------------------------------------------------
// Allocates a complex matrix, with row and column indices in the range
// [imin,imax] x [jmin,jmax]
//------------------------------------------------------------------------------
{                                                 // numbers of rows and columns
   int i, ni = imax - imin + 1, nj = jmax - jmin + 1;
   dcmplx **p;
                                               // allocate array of row pointers
   p = (dcmplx**)malloc((size_t)(ni * 2 * sizeof(double*)));
   if (!p) {
      printf("CMatrix: level 1 allocation error !\n");
      exit(1);
   }
   p -= imin;                                           // adjust for row offset
                                        // assign block start to 1st row pointer
   p[imin] = (dcmplx*)malloc((size_t)(ni*nj * 2 * sizeof(double)));
   if (!p[imin]) {
      printf("CMatrix: level 2 allocation error !\n");
      exit(2);
   }
   p[imin] -= jmin;                  // adjust 1st row pointer for column offset
                                     // define row pointers spaced by row length
   for (i = imin + 1; i <= imax; i++) p[i] = p[i - 1] + nj;

   return p;
}

//==============================================================================
void FreeCMatrix(dcmplx **p, int imin, int jmin)
//------------------------------------------------------------------------------
// Deallocates a complex matrix allocated with CMatrix, with row and column
// offsets imin and jmin
//------------------------------------------------------------------------------
{
   free((void*)(p[imin] + jmin));                            // deallocate block
   free((void*)(p + imin));                  // deallocate array of row pointers
}

//==============================================================================
char *String(int nchar)
//------------------------------------------------------------------------------
// Allocates a string of nchar characters
//------------------------------------------------------------------------------
{
   char *p;
                                          // assign block start to array pointer
   p = (char*)malloc((size_t)((nchar + 1) * sizeof(char)));
   if (!p) { printf("String: allocation error !\n"); exit(1); }
}

//==============================================================================
void FreeString(int *p)
//------------------------------------------------------------------------------
// Deallocates a string allocated with String
//------------------------------------------------------------------------------
{
   free((void*)p);
}

//==============================================================================
char **Strings(int n, int nchar)
//------------------------------------------------------------------------------
// Allocates a vector of n \0-terminated strings of nchar characters
//------------------------------------------------------------------------------
{
   int i;
   char **p;
                                            // allocate array of string pointers
   p = (char**)malloc((size_t)(n * sizeof(char*)));
   if (!p) { printf("Strings: level 1 allocation error !\n"); exit(1); }
                                     // assign block start to 1st string pointer
   p[0] = (char*)malloc((size_t)(n*(nchar + 1) * sizeof(char)));
   if (!p[0]) { printf("Strings: level 2 allocation error !\n"); exit(2); }
                      // define string pointers spaced by string length + 1 (\0)
   for (i = 1; i < n; i++) p[i] = p[i - 1] + nchar + 1;

   return p;
}

//==============================================================================
void FreeStrings(int **p)
//------------------------------------------------------------------------------
// Deallocates an int matrix allocated with IMatrix, with row and column
// offsets imin and jmin
//------------------------------------------------------------------------------
{
   free((void*)p[0]);                                        // deallocate block
   free((void*)p);                           // deallocate array of row pointers
}
